{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
--
-----------------------------------------------------------------------------

module Data.HDB.NormalFormsHUnit where

import Data.HDB.CoddFish ()
import Data.HDB.FunctionalDependencies
import Data.HDB.NormalForms
import Data.HDB.Representation
import Data.HListExtras
import Test.HUnit

-----------------------------------------------------------------------------

main
  = runTestTT . TestList $ testList
  
labeledTest l t
  = TestLabel l (TestCase t)

testList
  = [
    labeledTest "2ndNF" (
        do 
           assertEqual "Result should be False" False (booleanValue $ is2ndNF h1 fds)),
    labeledTest "Bcodd" (
           assertEqual "Result should be False" False (booleanValue $ boyceCoddNF all1 fds2))]

-- Examples
data NUMB
numb = attribute :: (Attribute Int NUMB) 

instance Show (Attribute Int NUMB) where show _ = "NUMB"

name = (attribute ) :: (Attribute String One) 
mark = (attribute ) :: (Attribute Float Two) 
coursecode =  (attribute ) :: (Attribute String Three)
coursename =  (attribute ) :: (Attribute String Four)

h1 = numb .*. name .*. mark .*. coursecode .*. HNil
h1'' = coursecode .*. coursename .*. HNil
h1'= name .*. mark .*. coursecode .*. coursename .*. HNil
newh = mark .*. coursecode .*. coursename .*. HNil 

fds = (FD (numb .*. HNil) (name .*. HNil)) .*. (FD (numb .*. coursecode .*. HNil) (mark .*. HNil)) .*. HNil
fds' = (FD (numb .*. HNil) (name .*. HNil)) .*. (FD (coursecode .*. HNil) (coursename .*. HNil)) .*. HNil
fds'' = (FD (coursecode .*. HNil) (coursename .*. HNil)) .*. HNil

-- Example 7.5 page 388 Ullman
a = (attribute ) :: (Attribute Int ()) 
b = (attribute ) :: (Attribute Int HZero) 
c = (attribute ) :: (Attribute Int One) 
d = (attribute ) :: (Attribute Int Two) 
e = (attribute ) :: (Attribute Int Three) 
f = (attribute ) :: (Attribute Int Four) 
g = (attribute ) :: (Attribute Int Five) 

fds1 = (FD (a .*. b .*. HNil) (c .*. HNil)) .*. 
       (FD (c .*. HNil) (a .*. HNil)) .*. 
       (FD (b .*. c .*. HNil) (d .*. HNil)) .*. 
       (FD (a .*. c .*. d .*. HNil) (b .*. HNil)) .*. 
       (FD (d .*. HNil) (e .*. HNil)) .*. 
       (FD (d .*. HNil) (g .*. HNil)) .*. 
       (FD (b .*. e .*. HNil) (c .*. HNil)) .*. 
       (FD (c .*. g .*. HNil) (b .*. HNil)) .*. 
       (FD (c .*. g .*. HNil) (d .*. HNil)) .*. 
       (FD (c .*. e .*. HNil) (g .*. HNil)) .*.
       (FD (c .*. e .*. HNil) (a .*. HNil)) .*. HNil

x = b .*. d .*. HNil 
u = a .*. b .*. c.*.  d.*.  e.*.  f .*. g .*. HNil



-- Example 7.10/7.12 page 398 Ullman

ct = (attribute ) :: (Attribute String ()) 
s = (attribute ) :: (Attribute String HZero) 
z = (attribute ) :: (Attribute String One) 

all1 = ct .*. s .*. z .*. HNil
fds2 = (FD (ct .*. s .*. HNil) (z .*. HNil)) .*.
       (FD (z .*. HNil) (ct .*. HNil)) .*.  HNil

-- Example 7.13 page 402 Ullman

a' = (attribute ) :: (Attribute String ()) 
b' = (attribute ) :: (Attribute String HZero) 
c' = (attribute ) :: (Attribute String One) 
d' =  (attribute ) :: (Attribute String Two) 

all2 = a' .*. b' .*. c' .*. d' .*. HNil
fds3 = (FD (a' .*. b' .*. HNil) (c' .*. HNil)) .*.
       (FD (c' .*. b' .*. HNil) (a' .*. HNil)) .*.
       (FD (b' .*. HNil) (d' .*. HNil)) .*.  HNil

--ks = keys all2 fds3

-- Example 7.14 page 402 Ullman
 
s' = (attribute ) :: (Attribute String ()) 
a'' = (attribute) :: (Attribute String HZero) 
i' = (attribute ) :: (Attribute String One) 
p' = (attribute ) :: (Attribute String Two) 

h = s' .*. a'' .*. i' .*. p' .*. HNil 

fds4 = (FD (s' .*. i' .*. HNil) (p' .*. HNil)) .*.
       (FD (s' .*. HNil) (a'' .*. HNil)) .*. HNil

