{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- 
--
-----------------------------------------------------------------------------

module Data.HDB.Read where

import Data.HListExtras
import Data.HDB.Representation
import Data.HDB.FunctionalDependencies
import Data.HDB.Advanced ()
import Data.HDB.Databases
import Language.Gnumeric.Syntax
import Language.Gnumeric.Parser (readXml')
import qualified Data.Map as Map
import GHC.List as List
import Text.XML.HaXml.Xml2Haskell hiding (Default)


----
import StrategyLib
import System
---


-- * Read Gnumeric File

-- | reads Gnumeric File to a database : 
        -- Each sheet in the workbook is supposed to have a single table 
        -- with the first row being the header (and therefore ignored)
        -- Referential integrety of resulting database is checked

readRDB fp h fds tns fks = do (wb ::  Gmr'Workbook) <- readGnumeric' fp
                              let lCells = map (tail.readTable.cells) $ sheets wb 
                              return $ DB $ Record $ hZip tns $ hZip (castTables h fds lCells) fks
                      

-- Maps a list of Cells to a list of Values (Strings) in the rigth position (Row,Collumn)    
readTable (Gmr'Cells cells) = Map.elems (Map.map Map.elems (foldr worker Map.empty cells))
                   where worker (Gmr'Cell ats value) m = let (i :: Int) = read $ gmr'CellRow ats
                                                             (j :: Int) = read $ gmr'CellCol ats
                                                         in Map.insertWith Map.union i (Map.singleton j value) m

-- Cast a matrix of strings into a list of tables
class CastTables h fds t | h fds -> t where
    castTables :: h -> fds -> [[[String]]] -> t

instance CastTables HNil fds HNil where
    castTables _ _ _ = HNil

instance (
            CastTables hs fds' ts, 
            CastTable (a,b) fds (Table (a,b) k v)
          ) 
       => CastTables (HCons (a,b) hs) (HCons fds fds') (HCons (Table (a,b) k v) ts) where
    castTables (HCons (a,b) hs) (HCons fds fds') (t:ts) = HCons (castTable (a,b) fds t) (castTables hs fds' ts)

-- | Class that allows the transformation of a matrix of Strings to a table according a given header
     -- When the table is not according to fds it returns Nothing 
     -- there was no need to have this function inside a class. however to avoid huge constraint headers 
     -- I've used this class to gather some of teh constraints
class CastTable h fds t | h -> t where
    castTable :: h -> fds -> [[String]] -> t

instance (CheckFDs fds (Table (a,b) k v), HeaderFor (a,b) k v, CastRow (a,b) (k,v)) => CastTable (a,b) fds (Table (a,b) k v) where
    castTable (a,b) fds rows | checkFDs fds t = t
                             | otherwise = Table (a,b) Map.empty
            where t = Table (a,b) (Map.fromList $ map (castRow (a,b)) rows) 

-- | cast of a single row
class CastRow h v | h -> v where
    castRow :: h -> [String] -> v

instance CastRow HNil HNil where
    castRow _ _ = HNil

instance (CastRow l vs, Read v) => CastRow (HCons (Attribute v nr) l) (HCons v vs) where
    castRow (HCons at ats) (s:ss) = HCons (read s) (castRow ats ss)

instance (CastRow l vs, Read v) => CastRow (HCons (NullableAttribute v nr) l) (HCons (Maybe v) vs) where
    castRow (HCons at ats) (s:ss) | s == "" = HCons Nothing (castRow ats ss)
                                  | otherwise = HCons (Just (read s)) (castRow ats ss)

instance (CastRow l vs, Read v) => CastRow (HCons (DefaultableAttribute v nr) l) (HCons v vs) where
    castRow (HCons (Default v) ats) (s:ss) | s == "" = HCons v (castRow ats ss)
                                           | otherwise = HCons (read s) (castRow ats ss)

-- | Lifting to pairs
instance (CastRow a k, CastRow b v, HLength a n, HNat2Integral n) => CastRow (a,b) (k,v) where
    castRow (a,b) l = let (k,v) = splitAt (hNat2Integral $ hLength a) l
                      in (castRow a k, castRow b v) 

---------------------
-- * Auxiliary stuff for gnumeric

-- | Reads a gnumeric file into a workbook
readGnumeric' :: (Term a, XmlContent a) => FilePath -> IO a
readGnumeric' fileName = do
  let tmpFile = "/tmp/TestGnumericTEMPFILE"
  let systemCmd = "zcat \""++fileName++"\" > "++tmpFile
  system systemCmd
  content <- readFile tmpFile
  let parseResult = {- parseFormulas $ -} readXml' content
  case (parseResult) of
    (Just workbook) -> return workbook
    Nothing -> error "NO PARSE"

-- | Selector for list stored in field sheets
sheets (Gmr'Workbook a b c d e f g h i (Gmr'Sheets sts) k l) = sts

-- | Selector for field cells
cells (Gmr'Sheet a b c d e f g h i j k l cs n o p q)
 = cs


