{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Unit tests.
--
-----------------------------------------------------------------------------

module Data.HDB.RelationalAlgebraHUnit where

import Data.HDB.Representation
import Data.HDB.RelationalAlgebra
import Data.HListExtras
import Prelude hiding (filter)
import qualified Data.Set as Set

import Test.HUnit

-----------------------------------------------------------------------------

main
  = runTestTT . TestList $ testList
  
labeledTest l t
  = TestLabel l (TestCase t)

testList
  = [
  labeledTest "testFilter" $ do
    assertEqual "empty" emptyTable (filter (const True) emptyTable)
    assertEqual "keep all" y (filter (const True) y)
    assertEqual "remove all" emptyTable (filter (const False) y)
    assertEqual "keep john" x (filter (hElem "John" . uncurry hAppend) y)
    assertEqual "keep 1" x (filter (hElem (1::Int) . uncurry hAppend) y)
  ,
  labeledTest "testProject" $ do
    assertEqual "empty" (ResultSet HNil Set.empty) (project HNil emptyTable)
    assertEqual "keep key only" 
      (ResultSet ((attribute :: Attribute Int ())   .*. HNil) 
             (Set.fromList [((1::Int) .*. HNil) ])) 
      (project ((attribute :: Attribute Int ())   .*. HNil) x)
    assertEqual "keep phone only" 
      (ResultSet ((attribute :: Attribute String Two)   .*. HNil) 
             (Set.fromList [("555-1234" .*. HNil) ])) 
      (project ((attribute :: Attribute String Two) .*. HNil) x)]
{-  ,
  labeledTest "test natural join" $ do
    let emptyRS = ResultSet HNil Set.empty
    assertEqual "empty" emptyRS (naturalJoin emptyRS emptyRS)
    assertEqual "empty" emptyRS (naturalJoin emptyRS (ResultSet HNil (Set.singleton HNil)))
    let r = naturalJoin
             (ResultSet employees (Set.singleton jane))
             (ResultSet employees (Set.singleton john))
    assertEqual "no rows" (ResultSet employees Set.empty) r
  ] -}
  
-- Examples

emptyTable = createTable (
  (attribute :: Attribute Int ())   .*. HNil
 ,
  (attribute :: Attribute String Zero)  .*.
  (attribute :: Attribute String One)   .*. 
  (attribute :: Attribute String Two)   .*. 
  HNil
 )

x = insert john emptyTable
y = insert jane x
     
john = (1::Int) .*. "John" .*. "Doe" .*. "555-1234" .*. HNil
jane = (0::Int) .*. "Jane" .*. "Doe" .*. "555-1234" .*. HNil

employees
 = (attribute  :: Attribute Int ())       .*.
   (attribute  :: Attribute String Zero)  .*.
   (attribute  :: Attribute String One)   .*. 
   (attribute  :: Attribute String Two)   .*. 
   HNil


