{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra Silva and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- 
--
-----------------------------------------------------------------------------

module Data.HDB.Write where

import Data.HListExtras
import Data.HDB.Representation
import Language.Gnumeric.Syntax
import qualified Data.Map as Map
import GHC.List as List


-- * Write CoddFish Table to Gnumeric Workbook

writeRDB db = addSheetNameIndex sheetNameIndex $ addSheets sheets emptywb
        where emptywb = emptyWorkbook
              (sheetNames, shts) = writeTables db
              sheets = Gmr'Sheets shts
              sheetNameIndex = Just $ Gmr'SheetNameIndex sheetNames

              

-- | writeTables collects  the sheet names

class WriteTables l l' | l -> l' where
    writeTables :: l -> ([Gmr'SheetName], l')
instance WriteTables (Record HNil) [Gmr'Sheet] where 
    writeTables _ = ([],[])

instance (
          Show l,
          WriteTables (Record ts) [Gmr'Sheet],
          WriteTable (a,b) ks vs [Gmr'Cell]
         )
          => WriteTables (Record (HCons (l,(Table (a, b) ks vs,fks)) ts)) [Gmr'Sheet] where
    writeTables (Record (HCons (l,(t,_)) ts)) = ((Gmr'SheetName name) : names, (writeTable name t) : ts')
        where name = show l
              (names, ts') = writeTables (Record ts)

type TableName = String
-- | Write a single table to a Gmr'WorkSheet
writeTable :: WriteTable (a,b) ks vs [Gmr'Cell] =>
                 TableName -> Table (a, b) ks vs -> Gmr'Sheet
writeTable s (Table (a,b) m) = addCells cells emptysht
        where emptysht = emptySheet s
              cells = Gmr'Cells (header++cs)
              header = writeRow 0 0 (a,b)
              cs = worker 1 0 (Map.toList m)
              worker i j l | (List.null l) = []
                           | otherwise = (writeRow i j (head l)) ++ (worker (i+1) j (tail l))

class (
       WriteRow (ks, vs) l,
       WriteRow h l, 
       HeaderFor h ks vs
     ) => WriteTable h ks vs l 
instance (
       WriteRow (ks, vs) [Gmr'Cell],
       WriteRow (a, b) [Gmr'Cell], 
       HeaderFor (a, b) ks vs
     ) => WriteTable (a,b) ks vs [Gmr'Cell]

class WriteRow l l' | l -> l' where
-- writeRow :: Row -> Collumn -> List of Values -> Gmr'Cell
    writeRow :: Int -> Int -> l -> l'
instance WriteRow HNil [Gmr'Cell] where 
    writeRow _ _ _ = []
instance (WriteRow l [Gmr'Cell], Show e) => WriteRow (HCons e l) [Gmr'Cell] where 
    writeRow r c (HCons e l) = (cell : (writeRow r (c+1) l))
                where cell = Gmr'Cell attrs (show e) 
                      attrs = (Gmr'Cell_Attrs (show c) (show r) Nothing Nothing Nothing Nothing Nothing Nothing) 
instance (WriteRow l [Gmr'Cell], Show e) => WriteRow (HCons (Maybe e) l) [Gmr'Cell] where 
    writeRow r c (HCons e l) = (cell : (writeRow r (c+1) l))
                where cell = case e of 
                                Just x -> Gmr'Cell attrs (show x) 
                                _      -> Gmr'Cell attrs "" 
                      attrs = (Gmr'Cell_Attrs (show c) (show r) Nothing Nothing Nothing Nothing Nothing Nothing) 
instance (HAppend a b ab, WriteRow ab [Gmr'Cell]) => WriteRow (a,b) [Gmr'Cell] where
    writeRow i j (a,b) = writeRow i j (hAppend a b)

-- * Auxiliary stuff for gnumeric

-- | empty workbook 
emptyWorkbook = Gmr'Workbook attrs version attributes summary
                 convention sheetNameIndex names
                 style geometry sheets uIData
                 calculation
        where attrs = Gmr'Workbook_Attrs (Just "http://www.gnumeric.org/v10.dtd") 
                                         (Just "http://www.w3.org/2001/XMLSchema-instance") 
                                         (Just "http://www.gnumeric.org/v8.xsd")
              version = Just $ Gmr'Version (Just "1") (Just "4") (Just "3") (Just "1.4.3")
              attributes = Just (Gmr'Attributes [Gmr'Attribute (Just (Gmr'type "4")) (Just (Gmr'name "WorkbookView::show_horizontal_scrollbar")) (Just (Gmr'value "TRUE")),Gmr'Attribute (Just (Gmr'type "4")) (Just (Gmr'name "WorkbookView::show_vertical_scrollbar")) (Just (Gmr'value "TRUE")),Gmr'Attribute (Just (Gmr'type "4")) (Just (Gmr'name "WorkbookView::show_notebook_tabs")) (Just (Gmr'value "TRUE")),Gmr'Attribute (Just (Gmr'type "4")) (Just (Gmr'name "WorkbookView::do_auto_completion")) (Just (Gmr'value "TRUE")),Gmr'Attribute (Just (Gmr'type "4")) (Just (Gmr'name "WorkbookView::is_protected")) (Just (Gmr'value "FALSE"))])
              summary = Just (Gmr'Summary [Gmr'Item (Gmr'name "application") (Just (Gmr'val_string "gnumeric")) Nothing,Gmr'Item (Gmr'name "application") (Just (Gmr'val_string "gnumeric")) Nothing]) 
              convention = []
              sheetNameIndex = Just $ Gmr'SheetNameIndex []
              names = []
              style = Nothing
              geometry = Gmr'Geometry {gmr'GeometryWidth = "768", gmr'GeometryHeight = "387"}
              uIData = Just (Gmr'UIData {gmr'UIDataSelectedTab = "0"}) 
              calculation = Just (Gmr'Calculation {gmr'CalculationManualRecalc = Just "0", gmr'CalculationEnableIteration = Just "1", gmr'CalculationMaxIterations = Just "100", gmr'CalculationIterationTolerance = Just "0.001"})
              sheets = Gmr'Sheets []


-- | empty sheet 
emptySheet s = Gmr'Sheet attrs name maxCol maxRow zoom names
               printInfo styles cols rows selections objects
               cells mergedRegions sheetLayout solver scenarios
        where attrs = Gmr'Sheet_Attrs "0" "0" "0" "0" "0" "1" "1" "1" Nothing
              name  = Gmr'Name s
              maxCol = Just (Gmr'MaxCol "-1")
              maxRow = Just (Gmr'MaxRow "-1")
              zoom = Gmr'Zoom "1"
              names = []
              j120 = Just "120"
              j0= Just "0.000000"
              jpt = Just "Pt"
              jcm = Just "cm"
              printInfo  = Just (Gmr'PrintInformation (Gmr'Margins (Gmr'top j120 jcm) (Gmr'bottom j120 jcm) (Gmr'left j0 jpt) (Gmr'right j0 jpt) (Gmr'header j0 jpt) (Gmr'footer j0 jpt)) (Gmr'Scale "percentage" (Just "100") Nothing Nothing) (Gmr'vcenter "0") (Gmr'hcenter "0") (Gmr'grid "0") (Gmr'even_if_only_styles "0") (Gmr'monochrome "0") (Gmr'draft "0") (Gmr'titles "0") Nothing Nothing (Gmr'order "d_then_r") (Gmr'orientation "portrait") (Gmr'Header "" "&amp;[TAB]" "") (Gmr'Footer "" "Page &amp;[PAGE]" "") Nothing)
              styles = Gmr'Styles sts
              sts = [Gmr'StyleRegion (Gmr'StyleRegion_Attrs "0" "0" "255" "65535") (Gmr'Style (Gmr'Style_Attrs (Just "1") (Just "2") (Just "0") Nothing (Just "0") (Just "0") (Just "0:0:0") (Just "FFFF:FFFF:FFFF") (Just "0:0:0") (Just "1") (Just "0") "General" (Just "0") (Just "0")) (Gmr'Font (Gmr'Font_Attrs "10" "0" "0" "0" "0") "Sans") (Just (Gmr'StyleBorder (Gmr'Top "0" Nothing) (Gmr'Bottom "0" Nothing) (Gmr'Left "0" Nothing) (Gmr'Right "0" Nothing) (Gmr'Diagonal "0" Nothing) (Gmr'Rev_Diagonal "0" Nothing))))]
              cols = Gmr'Cols (Gmr'Cols_Attrs "48") []
              rows = Gmr'Rows (Gmr'Rows_Attrs "12.75") []
              selections =  Gmr'Selections (Gmr'Selections_Attrs "0" "0") [Gmr'Selection {gmr'SelectionStartCol = "0", gmr'SelectionStartRow = "0", gmr'SelectionEndCol = "0", gmr'SelectionEndRow = "0"}] 
              objects = Nothing
              cells = Gmr'Cells []
              mergedRegions = Nothing 
              sheetLayout = Gmr'SheetLayout (Gmr'SheetLayout_Attrs "A1") Nothing
              solver = [Gmr'Solver (Gmr'Solver_Attrs {gmr'SolverTargetCol = Nothing, gmr'SolverTargetRow = Nothing, gmr'SolverProblemType = "1", gmr'SolverInputs = "", gmr'SolverMaxTime = Just "0", gmr'SolverMaxIter = Just "0", gmr'SolverNonNeg = Just "1", gmr'SolverDiscr = Just "0", gmr'SolverAutoScale = Just "0", gmr'SolverShowIter = Just "0", gmr'SolverAnswerR = Just "0", gmr'SolverSensitivityR = Just "0", gmr'SolverLimitsR = Just "0", gmr'SolverPerformR = Just "0", gmr'SolverProgramR = Just "0"}) []]
              scenarios = []

-- | Change SheetNameIndex field in workbook
addSheetNameIndex sNI (Gmr'Workbook a b c d e sNI' g h i j k l) = Gmr'Workbook a b c d e sNI g h i j k l

-- | Change sheets field in workbook
addSheets sheets (Gmr'Workbook a b c d e f g h i sheets' k l) = Gmr'Workbook a b c d e f g h i sheets k l

-- | Change cells field in sheet
addCells cells (Gmr'Sheet a b c d e f g h i j k l cells' n o p q)
 = Gmr'Sheet a b c d e f g h i j k l cells n o p q



