{-# OPTIONS -fglasgow-exts #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Parse Gnumeric spreadsheets.
--
----------------------------------------------------------------------------

module Language.Gnumeric.Parser where

import Language.Gnumeric.Syntax
import Language.Gnumeric.SyntaxTermInstances()
import Language.Gnumeric.Formula
import Language.Gnumeric.FormulaTermInstances()

import Text.XML.HaXml.Xml2Haskell
import Text.XML.HaXml.Escape
import Text.XML.HaXml.Types
import Text.XML.HaXml.Parse  (xmlParse)

import System
import StrategyLib
import Control.Monad.Identity
import Debug.Trace

import Text.PrettyPrint.HughesPJ (render)
import Text.XML.HaXml.Pretty (content)


-----------------------------------------------------------------------------

-- | Read, parse, and show a Gnumeric file.
testGnumeric :: IO ()
testGnumeric = do
  [fileName] <- getArgs
  (workbook::Gmr'Workbook) <- readGnumeric fileName
  putStrLn $ showXml workbook
  
-- | Read and parse a Gnumeric file.  
readGnumeric :: (Term a, XmlContent a) => FilePath -> IO a
readGnumeric fileName = do
  let tmpFile = "/tmp/TestGnumericTEMPFILE"
  let systemCmd = "zcat \""++fileName++"\" > "++tmpFile
  system systemCmd
  content <- readFile tmpFile
  let parseResult = parseFormulas $ readXml' content
  case (parseResult) of
    (Just workbook) -> return workbook
    Nothing -> error "NO PARSE"

-- | Alternative to 'readXml' that invokes 'xmlUnEscape' to transform XML
--   character references back to the charaters they represent.
readXml' :: XmlContent a => String -> Maybe a
readXml' s =
    let (Document _ _ y) = xmlParse "string input" s in
    fst (fromElem [CElem (xmlUnEscape stdXmlEscaper y)])

fst' (Nothing,rest)
  = error $ "REST: "++(concatMap (render.content) rest)
fst' (a,rest) 
  = a

-- | Parse the formulas in a Gnumeric spreadsheet or a fragment of a spreadsheet.
--   The parse results are stored in the AST attribute of each cell.
parseFormulas :: Term a => a -> a
parseFormulas a
  = maybe a id $ applyTP (stop_tdTP (failTP `adhocTP` worker)) a
    where
      worker :: Gmr'Sheet -> Maybe Gmr'Sheet
      worker c@(Gmr'Sheet _ (Gmr'Name name) _ _ _ _ _ _ _ _ _ _ cells _ _ _ _)
        = return $ parseFormulasOnSheet name c
  
-- | Parse all the formulas inside this sheet (a)
parseFormulasOnSheet :: Term a => String -> a -> a
parseFormulasOnSheet sheetName a
  = runIdentity $ applyTP (full_tdTP (idTP `adhocTP` worker)) a
    where
      worker :: Gmr'Cell -> Identity Gmr'Cell
      worker c@(Gmr'Cell _ "") = return c
      worker (Gmr'Cell attrs content) = do
        let f = parseFormula sheetName (mkCellId attrs) content
        let attrs' = attrs{ gmr'CellAst = Just f }
        return $
          trace ("Parsed formula: "++content
                 ++" at "++(mkCellId attrs)
                 ++" as: "++(show f))
            (Gmr'Cell attrs' content)


-- | Create the letter-number identifier for a cell, based on the row and column
--   numbers stored in the cell abstract syntax.
mkCellId :: Gmr'Cell_Attrs -> String
mkCellId attrs
  = col++row
    where
      rowNr = gmr'CellRow attrs
      colNr = gmr'CellCol attrs
      col = fromColNr (read colNr)
      row = fromRowNr (read rowNr)

-----------------------------------------------------------------------------

